/* simple NFA representation
 * pesco 2021-2022, isc license
 */

#include "nfa.h"
#include <stdio.h>
#include <ctype.h>
#include <assert.h>


/* replaceable default allocator */

NFA nfa_alloc(int);		/* must not return on error */

#ifndef NOALLOC

#include <stdint.h>	/* SIZE_MAX */
#include <err.h>

NFA
nfa_alloc(int sz)
{
	struct state *st;

	assert (sz >= 0);
	assert (sz <= SIZE_MAX / sizeof(struct state));
	st = malloc(sz * sizeof(struct state));
	if (sz > 0 && st == NULL)
		err(100, "nfa_alloc");

	return (NFA){st, sz};
}

#endif	/* NOALLOC */


/* elementary NFAs */

static NFA
state_(int x, int y, int cont)
{
	NFA a;

	a = nfa_alloc(1);
	a.state[0].chr = (struct range){x, y - x};
	a.state[0].edge_chr = cont;
	a.state[0].edge_eps = 0;		/* not present */

	return a;
}

NFA
fail(void)
{
	return state_(0, 0, 0);			/* this is going nowhere */
}

static NFA
range_(int x, int y)				/* [x, y) */
{
	return state_(x, y, 1);			/* continue to final state */
}

NFA
range(unsigned char min, unsigned char max)
{
	assert (min <= max);
	return range_(min, max + 1);
}

NFA
nrange(unsigned char min, unsigned char max)
{
	assert (min <= max);
	return range_(max + 1, min);		/* negative range */
}
	
NFA
chr(unsigned char x)
{
	return range(x, x);
}

NFA
any(void)
{
	return range_(0, OUTPUT);		/* any input symbol */
						/* i.e. everything up to END */
}

NFA
symbol(int x)					/* input only */
{
	assert (x >= 0);
	assert (x < OUTPUT);
	return range_(x, x + 1);
}

NFA
output(int x)
{
	assert (x >= OUTPUT);
	assert (x < INT_MAX);
	return range_(x, x + 1);
}

NFA
outchr(unsigned char x)
{
	return output(OUTPUT + x);
}

NFA
end(void)
{
	return symbol(END);
}

NFA
nop(void)
{
	/*
	 * the "null" range [0, 0) contains no element, but neither does its
	 * inverse (which is itself). we take this to mean that no input should
	 * be consumed and consider edge_chr an epsilon edge in this case.
	 *
	 * thus we can save the edge_eps field for use with choice().
	 * viewed another way, nop() can be inserted anywhere in the NFA
	 * "program" to add more epsilon edges to a state.
	 */
	return range_(0, 0);
}

NFA
think(int x)
{
	/*
	 * empty ranges with a base element different from 0 are used for
	 * "inner input". they behave like an epsilon edge iff the associated
	 * callback function returns true, i.e. tells us it thinks the edge
	 * should be active. this way these edges are easily handled by the
	 * iteration in epsilon_closure() below.
	 */
	assert (x >= INNER);
	assert (x < 0);
	return range_(x, x);
}

NFA
epsilon(void)
{
	return nfa_alloc(0);
}


/* combinators */

#include <stdarg.h>
#include <string.h>	/* strlen, memcpy */

NFA
str(const char *s)
{
	size_t i, len;
	NFA a;

	if (s[0] == '\0')
		return epsilon();

	len = strlen(s);
	if (len > INT_MAX)
		errx(1, "str: string longer than INT_MAX");
	a = nfa_alloc((int)len);
	for (i = 0; i < len; i++) {
		a.state[i].chr = (struct range){s[i], 1};
		a.state[i].edge_chr = 1;	/* next/final */
		a.state[i].edge_eps = 0;	/* not present */
	}
	
	return a;
}

NFA
seq_(NFA a, ...)
{
	va_list ap;
	int sz = 0;
	struct state *p;
	NFA b, sq;

	if (a.size < 0)
		return epsilon();

	/* determine total size of the sequence */
	va_start(ap, a);
	for (b = a; b.size >= 0; b = va_arg(ap, NFA)) {
		if (sz > INT_MAX - b.size)
			errx(1, "seq: would exceed INT_MAX");
		sz += b.size;
	}
	va_end(ap);

	/*
	 * concatenating the argument NFAs forms the sequence.
	 *
	 * note that each NFA's final edge points one past the end, so the
	 * virtual final state is fused with the next NFA's start state by
	 * construction.
	 */
	sq = nfa_alloc(sz);
	p = sq.state;
	va_start(ap, a);
	for (b = a; b.size >= 0; b = va_arg(ap, NFA)) {
		memcpy(p, b.state, b.size * sizeof(struct state));
		p += b.size;
	}
	va_end(ap);

	return sq;
}

NFA
choice_(NFA a, ...)
{
	va_list ap;
	int i, j, n = 0, sz = 0;
	struct state *p;
	NFA b, sq;

	/*
	 * choice() operates much like seq() but the out edges of each branch
	 * must be adjusted to point to the new end. also, an initial chain of
	 * nonconsuming "nop" states is generated, fanning out to each branch.
	 */
	// XXX Q: can we ensure always having a leading free eps?
	//     A: no, because the first state could have epsilon in-edges and
	//        that means we can't just add an epsilon out-edge.

	if (a.size < 0)
		return fail();

	/* determine total size of the result */
	va_start(ap, a);
	for (b = a; b.size >= 0; b = va_arg(ap, NFA)) {
		if (sz > INT_MAX - b.size - 1)
			errx(1, "choice: would exceed INT_MAX");
		sz += b.size;
		n++;
	}
	va_end(ap);
	sz += n - 1;	/* plus space for n - 1 forks */

	/* allocate result */
	sq = nfa_alloc(sz);
	p = sq.state;

	/* create a sequence of epsilon edges to each branch (fan out) */
	for (i = 0; i < n - 1; i++, p++) {
		p->chr = (struct range){0, 0};
		p->edge_chr = 1;
		p->edge_eps = 0;	/* filled below */
	}

	/* for all but the last arg: copy states, adjust out edges, add fork */
	va_start(ap, a);
	for (i = 0; i < n - 1; i++) {
		for (j = 0; j < a.size; j++, p++) {
			/* copy state */
			*p = a.state[j];

			/* adjust final edges to the new end */
			assert (p - sq.state < sz);
			if (j + p->edge_chr == a.size)
				p->edge_chr = sz - (p - sq.state);
			if (j + p->edge_eps == a.size)
				p->edge_eps = sz - (p - sq.state);
		}

		/* fork */
		sq.state[i].edge_eps = p - sq.state - i;

		/* next argument */
		a = va_arg(ap, NFA);
	}
	va_end(ap);

	/* for the last argument, just copy states */
	memcpy(p, a.state, a.size * sizeof(struct state));

	return sq;
}

NFA
opt(NFA a)
{
	return choice(a, epsilon());
}

NFA
rep(int n, NFA a)
{
	NFA r;
	struct state *p;

	if (n == 0)
		return epsilon();

	/* allocate result */
	assert (n > 0);
	assert (a.size >= 0);
	if (a.size > INT_MAX / n)
		errx(1, "rep: would exceed INT_MAX");
	r = nfa_alloc(n * a.size);

	/* create n copies of 'a' back-to-back, cf. seq() */
	p = r.state;
	for (; n > 0; n--) {
		memcpy(p, a.state, a.size * sizeof(struct state));
		p += a.size;
	}

	return r;
}

NFA
plus(NFA a)
{
	NFA r;

	/* allocate result */
	assert (a.size >= 0);
	if (a.size > INT_MAX - 1)
		errx(1, "plus: would exceed INT_MAX");
	r = nfa_alloc(a.size + 1);

	/* a copy of 'a' followed by an epsilon edge back to the start */
	memcpy(r.state, a.state, a.size * sizeof(struct state));
	r.state[a.size].chr = (struct range){0, 0};
	r.state[a.size].edge_chr = 1;
	r.state[a.size].edge_eps = -a.size;

	return r;
}

NFA
star(NFA a)
{
	return opt(plus(a));
}


/* diagnostic output: print an NFA as a sequence of states */
void
nfaprint(NFA a)
{
	struct state s;
	int i;

	for(i = 0; i < a.size; i++) {
		s = a.state[i];

#if 0
		/* a simple uniform display format for the raw state data */
		printf("%5d: [%2Xh, %2Xh) -> %d, e -> %d\n", i,
		    (unsigned)s.chr.base, (unsigned)s.chr.base + s.chr.n,
		    i + s.edge_chr, i + s.edge_eps);
#else
		/*
		 * a big mess of conditionals to make the possible cases easy
		 * to read and distinguish for hunams
		 */
		printf("%5d: ", i);
		if (s.chr.n == 0) {
			/* edge_chr is an epsilon or inner input edge */
			if (s.chr.base != 0)
				printf("<inner %d>", s.chr.base - INNER);
			else if (s.edge_chr != 1)
				printf("e");		/* epsilon */
			if (s.edge_chr != 1)
				printf(" -> %d", i + s.edge_chr);
		} else {
			if (s.chr.n < 0)		/* negative range */
				printf(")%2Xh, %2Xh(",
				    (unsigned)s.chr.base + s.chr.n,
				    (unsigned)s.chr.base - 1);
			else if (s.chr.n == 1) {	/* singleton */
				if (s.chr.base >= OUTPUT) {
					printf(">output %d<",
					    s.chr.base - OUTPUT);
				} else {
					printf("%2Xh", (unsigned)s.chr.base);
					if (isprint(s.chr.base))
						printf(" '%c'", s.chr.base);
					else if (s.chr.base == END)
						printf(" END");
				}
			} else
				printf("[%2Xh, %2Xh]", (unsigned)s.chr.base,
				    (unsigned)s.chr.base + s.chr.n - 1);
			if (s.edge_chr != 1)
				printf(" -> %d", i + s.edge_chr);
		}
		if (s.edge_eps != 0) {
			if (s.chr.n != 0 || s.edge_chr != 1)
				printf(", ");
			printf("e -> %d", i + s.edge_eps);
		} else if (s.chr.n != 0) {		/* no epsilon edges */
			if (s.edge_chr == 0)		/* no future */
				printf(" (fail!)");
		}
		putchar('\n');
#endif
	}

	/* show the virtual final state for completeness */
	printf("%5d: (final)\n", i);
}


#ifdef BITSET
typedef unsigned long bunch_t;	/* a bunch of bits */
#define BUNCH_BIT (sizeof(bunch_t) * CHAR_BIT)
#else
typedef char bunch_t;
#define BUNCH_BIT 1
#endif

struct set {
	bunch_t *bits;
	size_t n;		/* number of bits */
	size_t sz;		/* number of bunches */
};

static int
inset(const struct set *set, int x)
{
#ifdef BITSET
	return (set->bits[x / BUNCH_BIT] >> (x % BUNCH_BIT)) & 1;
#else
	return set->bits[x] != 0;
#endif
}

static void
setinsert(struct set *set, int x)
{
#ifdef BITSET
	set->bits[x / BUNCH_BIT] |= (1L << (x % BUNCH_BIT));
#else
	set->bits[x] = 1;
#endif
}

static void
setremove(struct set *set, int x)
{
#ifdef BITSET
	set->bits[x / BUNCH_BIT] &= ~(1L << (x % BUNCH_BIT));
#else
	set->bits[x] = 0;
#endif
}

#define FOREACHBIT(I,SET)	\
	for (I = 0; I < (SET)->n; I++) if (inset(SET, I))
#define FOREACHSTATE(A,I,SET)	\
	for (I = 0; I < (A).size; I++) if (inset(SET, I))

/* diagnostic output: print input symbol and state set */
static void
printstates(const struct set *set, int x)
{
#if 0
	size_t i;

	/* active states */
	FOREACHBIT(i, set)
		printf("%zu ", i);

	/* input */
	if (x == END)
		printf(" $$\n");
	else
		printf(" %c\n", x);
#endif
}


struct prep {
	NFA a;
	sub_t *sub;		/* the "subconscious" mind of the automaton */
	out_t *out;		/* output subroutine */
	void *env;		/* user-supplied environment for sub and out */

	struct set act;		/* set of "active" states */
	struct set next;	/* set of "next active" states */
	struct set eps;		/* which states carry epsilon edges */

	int pos;		/* current input position */
	int match;		/* length of longest match so far */

	int sz;
	bunch_t buf[];
};

void
nfareset(struct prep *pr)
{
	pr->act.n = pr->next.n = pr->a.size + 1;
	pr->act.sz = pr->next.sz = pr->sz;
	pr->act.bits = pr->buf;
	pr->next.bits = pr->buf + pr->sz;
	pr->pos = 0;
	pr->match = -1;		/* no match */

	memset(pr->act.bits, 0, pr->act.sz * sizeof(bunch_t));
	setinsert(&pr->act, 0);
}

struct prep *
nfaprep(NFA a, sub_t *sub, out_t *out, void *env)
{
	struct prep *pr;
	int setsz;

	/* allocate */
	assert (a.size <= SIZE_MAX - sizeof(bunch_t));	// XXX ?
	setsz = (a.size + BUNCH_BIT) / BUNCH_BIT;	/* a.size + 1 bits */
	pr = malloc(sizeof(struct prep) + 3 * setsz * sizeof(bunch_t));
	assert (pr != NULL);	// XXX

	pr->a = a;
	pr->sz = setsz;
	pr->sub = sub;
	pr->out = out;
	pr->env = env;

	nfareset(pr);

#if BITSET
	/* compute epsilon map (which states carry epsilon edges) */
	pr->eps.bits = pr->buf + 2 * setsz;
	pr->eps.sz = setsz;
	pr->eps.n = a.size + 1;
	memset(pr->eps.bits, 0, pr->eps.sz * sizeof(bunch_t));
	for (int i = 0; i < a.size; i++)
		if (a.state[i].chr.n == 0)
			setinsert(&pr->eps, i);
#endif

	return pr;
}

/*
 * perform one iteration of epsilon-expansion over act
 *
 * also handles inner input symbols. edges marked with inner input are not
 * technically epsilon edges but they act the same, i.e. we
 * nondeterministically follow them, if our "subconscious" callback tells us to
 * activate them.
 *
 * returns 1 iff any states remain in the 'work' set, 0 otherwise.
 */
static int
epsilon_step(struct prep *pr, struct set *act, struct set *work)
{
	NFA a = pr->a;
	struct state s;
	size_t k, x;
	int workleft = 0;

#ifdef BITSET
	for (int i = 0; i < work->sz; i++)
	 if (work->bits[i] & pr->eps.bits[i])
	  for (int j = 0; j < BUNCH_BIT && (k = i * BUNCH_BIT + j) < a.size; j++)
	   if ((work->bits[i] >> j) & 1)
#else
	FOREACHSTATE(a, k, work)
#endif
	{
		s = a.state[k];

		/*
		 * we only create states with zero or two epsilon edges, so
		 * only bother looking at edge_eps if edge_chr is also an
		 * epsilon edge.
		 */
		if (s.chr.n == 0) {
			/*
			 * edge_chr
			 *
			 * normally, the empty range (n = 0) also has base = 0,
			 * but inner inputs are also handled here, indicated by
			 * base != 0. in that case, sub() must be called and
			 * the edge is only active if it returns true.
			 *
			 * NB: these edges can only point forward so we will
			 *     still visit their target in this pass.
			 */
			if (s.chr.base == 0 ||
			    (pr->sub != NULL &&
			     pr->sub(s.chr.base, k, pr->pos, pr->env))) {
				x = k + s.edge_chr;
				setinsert(work, x);
				setinsert(act, x);
			}

			/*
			 * edge_eps
			 *
			 * these can point forward or backward. in the latter
			 * case, we have already passed their target and must
			 * work it with another iteration (workleft = 1).
			 */
			if (s.edge_eps != 0) {
				x = k + s.edge_eps;
				if (!inset(act, x)) {
					if (s.edge_eps < 0)
						workleft = 1;
					setinsert(work, x);
					setinsert(act, x);
				}
			}
		} else
			assert(s.edge_eps == 0);

		setremove(work, k);		/* done with this one */
	}
	setremove(work, a.size);		/* this fully clears work */

	return workleft;
}

static void
epsilon_closure(struct prep *pr, struct set *act, struct set *wrk)
{
	/* initial workset = initial active set */
	assert(wrk->sz == act->sz);
	memcpy(wrk->bits, act->bits, act->sz * sizeof(bunch_t));

	while (epsilon_step(pr, act, wrk));
	/* NB: wrk is now empty */
}

static void
markactive(NFA a, struct set *next, int x)
{
	struct state s;

	/*
	 * avoid marking chains of states that have only epsilon out edges.
	 * these are not actually relevant for matching input.
	 * to keep things simple and free of recursion, we only follow the
	 * chain along edge_chr and do mark states pointed to by edge_eps.
	 */
	while ((s = a.state[x]).chr.n == 0) {		/* skip forks */
		setinsert(next, x + s.edge_eps);	/* mark branches */
		x += s.edge_chr;			/* follow edge_chr */
		if (x == a.size)			/* stop at final */
			break;
	}
	setinsert(next, x);				/* mark end of chain */
}

static int
matchchr(NFA a, const struct set *act, struct set *next, int x)
{
	size_t i;
	struct state s;
	int match = 0;

#if 0	// XXX assume 'next' is already empty
	/* clear next-set */
	memset(next->bits, 0, next->sz * sizeof(bunch_t));
#endif

	/* match input against all active states, filling next-set */
	FOREACHSTATE(a, i, act) {
		s = a.state[i];
#if 0
		if ((s.chr.n > 0 &&	/* positive range */
			(x >= s.chr.base && x < s.chr.base + s.chr.n)) ||
		    (s.chr.n < 0 &&	/* inverted range */
			(x >= s.chr.base || x < s.chr.base + s.chr.n)))
#else
		/* NB: unsigned arithmetic, i.e. modulo UINT_MAX+1 */
		if ((unsigned)x - s.chr.base < s.chr.n)
#endif
		{
			match = 1;
			markactive(a, next, i + s.edge_chr);
		}
	}

	return match;	/* i.e. "is next non-empty" */
}

/*
 * try to produce an output symbol given the current active set.
 * also activate the corresponding target state in next.
 *
 * output is predicative, i.e. out() may reject a given range (return -1).
 * in that case we continue trying other possibilities. the first to succeed
 * wins.
 *
 * returns 0 on success or -1 if no output is possible.
 */
static int
generate(struct prep *pr, const struct set *act, struct set *next)
{
	NFA a = pr->a;
	struct range r;
	size_t k;
	int x, b;

	if (pr->out == NULL)
		return -1;				/* no output possible */

	FOREACHSTATE(pr->a, k, act) {
		r = pr->a.state[k].chr;

		/* we assume that output ranges always have base >= OUTPUT */
		if (r.base < OUTPUT)			/* not output? */
			continue;			/* skip this state */

		assert(r.n != 0);			/* not epsilon */
		assert(r.n < INT_MAX - r.base);		/* no overflow */
		assert(r.base + r.n >= OUTPUT);		/* all output */

		x = pr->out(r, k, pr->pos, pr->env);	/* attempt output */
		if (x != -1) {
			b = matchchr(a, act, next, x);	/* fill next */
			assert(b == 1);			/* x can only match */
			return 0;			/* success */
		}
	}

	return -1;					/* no output possible */
}

/*
 * run the automaton for one iteration, either consuming input, generating
 * output or failing.
 *
 * returns the number of input characters consumed or -1 if stuck.
 */
static int
nfastep(struct prep *pr, int x, struct set *act, struct set *next)
{
	NFA a = pr->a;
	bunch_t *tmp;
	int r = 0;					/* input consumed? */

	printstates(act, x);				/* diagnostic output */

	/*
	 * form epsilon closure of act - also handles inner input.
	 * uses, and incidentally clears, our 'next' set for its work.
	 */
	epsilon_closure(pr, act, next);

	if (inset(act, a.size))				/* final state? */
		pr->match = pr->pos;			/* save position */

	if (matchchr(a, act, next, x))			/* input match? */
		r = 1;
	else						/* else try output */
		r = generate(pr, act, next);	

	/* swap sets */
	tmp = act->bits;
	act->bits = next->bits;
	next->bits = tmp;

	return r;
}

int
nfacont(struct prep *pr, const char *input, size_t sz)
{
	int i, r, x;

	for(i = 0; i < sz; i += r) {
		x = (unsigned char)input[i];
		r = nfastep(pr, x, &pr->act, &pr->next);
		if (r == -1)			/* no match */
			break;
		pr->pos += r;
	}

	return pr->match;
}

int
nfaend(struct prep *pr)
{
	/* run until output ceases or END matches */
	while(nfastep(pr, END, &pr->act, &pr->next) == 0);
	return pr->match;
}

int
nfarun(struct prep *pr, const char *input, size_t sz)
{
	nfareset(pr);
	nfacont(pr, input, sz);
	nfaend(pr);
	return pr->match;
}

int
nfamatch(NFA a, sub_t *sub, out_t *out, void *env, const char *inp, size_t sz)
{
	struct prep *pr;
	int result;

	pr = nfaprep(a, sub, out, env);
	result = nfarun(pr, inp, sz);
	free(pr);

	return result;
}
